///////////////////////////////////////////////////////////////////////////////
// File: protocones.cpp                                                      //
// Description: source file for stable cones determination (Cstable_cones)   //
// This file is part of the SISCone project.                                 //
// For more details, see http://www.lpthe.jussieu.fr/~soyez/siscone          //
//                                                                           //
// Copyright (c) 2006 Gavin Salam and Gregory Soyez                          //
//                                                                           //
// This program is free software; you can redistribute it and/or modify      //
// it under the terms of the GNU General Public License as published by      //
// the Free Software Foundation; either version 2 of the License, or         //
// (at your option) any later version.                                       //
//                                                                           //
// This program is distributed in the hope that it will be useful,           //
// but WITHOUT ANY WARRANTY; without even the implied warranty of            //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             //
// GNU General Public License for more details.                              //
//                                                                           //
// You should have received a copy of the GNU General Public License         //
// along with this program; if not, write to the Free Software               //
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA //
//                                                                           //
// $Revision:: 128                                                          $//
// $Date:: 2007-01-18 20:07:57 +0100 (Thu, 18 Jan 2007)                     $//
///////////////////////////////////////////////////////////////////////////////

#include "protocones.h"
#include "defines.h"
#include <math.h>
#include <iostream>

namespace siscone{

using namespace std;

/**********************************************************************
 * Cstable_cones implementation                                       *
 * Computes the list of stable comes from a particle list.            *
 * This class does the first fundamental task of te cone algorithm:   *
 * it is used to compute the list of stable cones given a list        *
 * of particles.                                                      *
 **********************************************************************/
// default ctor
//--------------
Cstable_cones::Cstable_cones(){
  nb_tot = 0;
  hc = NULL;
}

// ctor with initialisation
//--------------------------
Cstable_cones::Cstable_cones(vector<Cmomentum> &_particle_list)
  : Cvicinity(_particle_list){

  nb_tot = 0;
  hc = NULL;
}

// default dtor
//--------------
Cstable_cones::~Cstable_cones(){
  if (hc!=NULL) delete hc;
}


/*
 * initialisation
 *  - _particle_list  list of particles
 *  - _n              number of particles
 *********************************************************************/
void Cstable_cones::init(vector<Cmomentum> &_particle_list){
  // check already allocated mem
  if (hc!=NULL){
    delete hc;
  }
  if (protocones.size()!=0)
    protocones.clear();

  // initialisation
  set_particle_list(_particle_list);
}
  

/*
 * compute stable cones.
 * This function really does the job i.e. computes
 * the list of stable cones (in a seedless way)
 *  - _radius:  radius of the cones
 * The number of stable cones found is returned
 *********************************************************************/
int Cstable_cones::get_stable_cones(double _radius){
  int p_idx;

  // check if everything is correctly initialised
  if (n_part==0){
    return 0;
  }

  R  = _radius;
  R2 = R*R;

  // allow hash for cones candidates
  hc = new hash_cones(n_part, R2);

  // browse all particles
  for (p_idx=0;p_idx<n_part;p_idx++){
    // step 0: compute the child list CL.
    //         Note that this automatically sets the parent P
    build_from_list(&plist[p_idx], 2.0*R);

    // special case: 
    //   if the vicinity is empty, the parent particle is a 
    //   stable cone by itself. Add it to protocones list.
    if (vicinity_size==0){
      protocones.push_back(*parent);
      continue;
    }

    // setp 1: initialise with the first cone candidate
    init_cone();

    do{
      // step 2: test cone stability for that pair (P,C)
      test_cone();

      // step 3: go to the next cone child candidate C
    } while (!update_cone());
  }

  return proceed_with_stability();
}


/*
 * initialise the cone.
 * We take the first particle in the angular ordering to compute 
 * this one
 * return 0 on success, 1 on error
 *********************************************************************/
int Cstable_cones::init_cone(){
  // init set child C
  centre = vicinity[0];
  child = centre->v;
  centre_idx = 0;

#ifdef USE_QUADTREE
  compute_cone_from_quadtree();
#else
  compute_cone_from_list();
#endif

  return 0;
}

/*
 * test cones.
 * We check if the cone(s) build with the present parent and child 
 * are stable
 * return 0 on success 1 on error
 *********************************************************************/
int Cstable_cones::test_cone(){
  Creference weighted_cone_ref;
  
  // depending on the side we are taking the child particle,
  // we test different configuration.
  // Each time, two configurations are tested in such a way that
  // all 4 possible cases (parent or child in or out the cone)
  // are tested when taking the pair of particle parent+child
  // and child+parent.

  // here are the tests entering the first series:
  //  1. check if the cone is already inserted
  //  2. check cone stability for the parent and child particles
  
  if (centre->side){
    // test when both particles are not in the cone
    // or when both are in.
    // Note: for the totally exclusive case, test emptyness before
    cone_client = cone;
    if (cone.ref.not_empty()){
      hc->insert(&cone_client, parent, child, false, false);
    }

    cone_client = cone;
    cone_client+= *parent + *child;
    hc->insert(&cone_client, parent, child, true, true);
  } else {
    // test when 1! of the particles is in the cone
    cone_client = cone + *parent;
    hc->insert(&cone_client, parent, child, true, false);

    cone_client = cone + *child;
    hc->insert(&cone_client, parent, child, false, true);
  }

  nb_tot+=2;

  return 0;
}



/*
 * update the cone
 * go to the next child for that parent and update 'cone' appropriately
 * return 0 if update candidate found, 1 otherwise
 ***********************************************************************/
int Cstable_cones::update_cone(){
  double pt;

  // get the next child and centre
  centre_idx++;
  if (centre_idx==vicinity_size)
    return 1;

  // update the cone w.r.t. the old child
  // only required if the old child is entering inside in which
  // case we need to add it. We also know that the child is 
  // inside iff its side is -.
  if (!centre->side){
    // update cone
    cone += (*child);

    // update stability check quantities
    pt = child->perp2();
    dpt += pt;
    dpt_abs += pt;
  }

  // update centre and child
  centre = vicinity[centre_idx];
  child = centre->v;

  // update the cone w.r.t. the new child
  // only required if the new child was already inside in which
  // case we need to remove it. We also know that the child is 
  // inside iff its side is +.
  if ((centre->side) && (cone.ref.not_empty())){
    // update cone
    cone -= (*child);

    // update stability check quantities
    pt = child->px*child->px + child->py*child->py;
    dpt -= pt;
    dpt_abs += pt;
  }

  // check that the addition and substraction of vectors does
  // not lead to too much rounding error
  // for that, we compute the sum of pt modifications and of |pt|
  // since last recomputation and once the ratio overpasses a threshold
  // we recompute vicinity.
  if (fabs(dpt)<PT_TSHOLD*dpt_abs){
#ifdef USE_QUADTREE
    compute_cone_from_quadtree();
#else
    compute_cone_from_list();
#endif
  }

  return 0; 
}


/*
 * compute the cone momentum using the particle list.
 * given parent, child and centre compute the momentum
 * of the particle inside the cone
 */
void Cstable_cones::compute_cone_from_list(){
  unsigned int i;
  double cx, cy, dx, dy;
  Cmomentum *v;

  // init cone (without P & C)
  // the cone contains all the particles within a radius R
  // and is computed from centre
  cx = centre->eta;
  cy = centre->phi;


  // set momentum to 0
  cone = Cmomentum();

  // Important note: we can only browse the particles
  // in vicinity since all particles in the cone are
  // withing a distance 2R w.r.t. parent hence in vicinity.
  for (i=0;i<vicinity_size;i++){
    // to avoid double-counting, only use particles with + angle
    if (vicinity[i]->side){
      v = vicinity[i]->v;
      if (v!=child){
	dx = cx - v->eta;
	dy = fabs(cy - v->phi);
	if (dy>M_PI) 
	  dy -= pi2;
	
	if (dx*dx+dy*dy<R2){
	  cone+=*v;
	}
      }
    }
  }

  // set check variables back to 0
  dpt = 0.0;
  dpt_abs = 0.0;
}


/*
 * compute the cone momentum using the quadtree.
 * given parent, child and centre compute the momentum
 * of the particle inside the cone
 */
void Cstable_cones::compute_cone_from_quadtree(){
  // store centre coordinates temporarily (pcx,pcy defined in Cvicinity)
  pcx = centre->eta;
  pcy = centre->phi;

  // init momentum to 0 and start with parent quadtree
  cone = Cmomentum();
  compute_cone_from_quadtree_internal(quadtree);

  // set check variables back to 0
  dpt = 0.0;
  dpt_abs = 0.0;
}


/*
 * compute the cone momentum using the quadtree.
 * given parent, child and centre compute the momentum
 * of the particle inside the cone
 * internal usage (called from compute_cone_from_quadtree)
 */
void Cstable_cones::compute_cone_from_quadtree_internal(Cquadtree *qt){
  // Description of the method:
  // --------------------------
  // 1. check if cell is empty => no intersection
  // 2. if cell has 1! particle, check if it is inside the circle.
  //    If yes, add it and return, if not simply return.
  // 3. check if the circle intersects the square. If not, return.
  // 4. check if the square is inside the circle. 
  //    If yes, add it to qt and return.
  // 5. check intersections with childs.
  //
  // Note: there is a theoretical failure risk here: if the parent || child
  //       is at the border of a fully included cell it can be included.
  //       This has to be the border since they are at the border of the circle

  Cmomentum *v = qt->v;

  // step 1: if there is no particle inside te square, no reason to go further
  if (v==NULL)
    return;

  double dx, dy;

  // step 2: if there is only one particle inside the square, test if it is in
  //         the circle, in which case add the square to 'qt', and return
  if (!qt->has_child){
    // compute the distance
    // Note: v has only one particle => eta and phi are defined
    dx = pcx - v->eta;
    dy = fabs(pcy - v->phi);
    if (dy>M_PI) 
      dy -= pi2;

    // test distance
    if (dx*dx+dy*dy<R2){
      // prevent from adding parent or child
      if ((v==parent) || (v==child))
	return;

      cone+=(*v);
    }

    return;
  }

  // step 3: check if there is an intersection
  double dx_c, dy_c;

  // compute (minimal) the distance (pay attention to the periodicity in phi).
  dx_c = fabs(pcx - qt->centre_x);
  dx =  dx_c - qt->half_size_x;
  if (dx<0.0) dx=0.0;

  dy_c = fabs(pcy - qt->centre_y);
  if (dy_c>M_PI)
    dy_c = pi2-dy_c;
  dy = dy_c - qt->half_size_y;
  if (dy<0.0) dy=0.0;

  // check the distance 
  if (dx*dx+dy*dy>=R2){
    // no intersection
    return;
  }

  // step 4: check if included

  // compute the (maximal) distance
  dx = dx_c + qt->half_size_x;
  
  dy = dy_c + qt->half_size_y;
  if (dy>M_PI) 
    dy = pi2-dy;

  if (dx*dx+dy*dy<R2){
    cone+=(*v);
    return;
  }

  // step 5: the square is not fully in. Recurse to childs
  compute_cone_from_quadtree_internal(qt->childs[0][0]);
  compute_cone_from_quadtree_internal(qt->childs[0][1]);
  compute_cone_from_quadtree_internal(qt->childs[1][0]);
  compute_cone_from_quadtree_internal(qt->childs[1][1]);
}


/*
 * compute stability of all enumerated clients.
 * For all client cones which are stable w.r.t. their border particles,
 * pass the last test: stability with quadtree intersection
 ************************************************************************/
int Cstable_cones::proceed_with_stability(){
  int i,n;
  hash_element *elm;

  n=0;
  for (i=0;i<=hc->mask;i++){
    // test ith cell of the hash array
    elm = hc->hash_array[i];

    // browse elements therein
    while (elm!=NULL){
      // test stability
      if (elm->is_stable){
	// stability is not ensured by all pairs of "edges" already browsed
#ifdef USE_QUADTREE
	//  => testing stability with quadtree intersection
	if (quadtree->circle_intersect(elm->eta, elm->phi, R2)==elm->ref)
#else
	//  => testing stability with the particle-list intersection
	if (circle_intersect(elm->eta, elm->phi)==elm->ref)
#endif
	  protocones.push_back(Cmomentum(elm->eta, elm->phi, elm->ref));
      }
      
      // jump to the next one
      elm = elm->next;
    }
  }
  
  // free hash
  // we do that at this level because hash eats rather a lot of memory
  // we want to free it before running the split/merge algorithm
  delete hc;
  hc=NULL;

  return protocones.size();
}

/*
 * circle intersection.
 * computes the intersection with a circle of given centre and radius.
 * The output takes the form of a checkxor of the intersection's particles
 *  - cx    circle centre x coordinate
 *  - cy    circle centre y coordinate
 * return the checkxor for the intersection
 ******************************************************************/
Creference Cstable_cones::circle_intersect(double cx, double cy){
  Creference intersection;
  int i;
  double dx, dy;

  for (i=0;i<n_part;i++){
    // compute the distance of the i-th particle with the parent
    dx = plist[i].eta - cx;
    dy = fabs(plist[i].phi - cy);
    
    // pay attention to the periodicity in phi !
    if (dy>M_PI) 
      dy -= pi2;
    
    // really check if the distance is less than VR
    if (dx*dx+dy*dy<R2)
      intersection+=plist[i].ref;
  }
  
  return intersection;
}
  
}
